//+------------------------------------------------------------------+
//|                                              EqualVolumeBars.mq5 |
//|                                 Copyright © 2008-2015, komposter |
//|                                      Copyright © 2018, Marketeer |
//|                        https://www.mql5.com/en/blogs/post/719419 |
//+------------------------------------------------------------------+
#property copyright "Copyright © 2018, Marketeer"
#property link "https://www.mql5.com/en/users/marketeer"
#property description "Non-trading EA generating equivolume and/or range bars as a custom symbol.\n"
#property description "This is a port of MT4 offline chart generator to MT5.\n"
#property description "Original MT4 EqualVolumeBars - Copyright © 2008-2015, www.komposter.me.\n"
#property version "5.0"

#include <Symbol2.mqh>

// comment on/off this include depending from build: builds 1860+ go with native support of timeseries functions
// iTime, iOpen, iClose, iHigh, iLow, iVolume
//#include <mt4_dataseries.mqh>

#define TICKS_ARRAY 10000

enum mode
{
  EqualVolumeBars = 0,
  RangeBars = 1
  // TODO: EqualRealVolumeBars = 2
};

input mode WorkMode = EqualVolumeBars;
input int TicksInBar = 100;
input bool FromM1 = true;
input datetime StartDate = 0;
input string CustomPrefix = "";

datetime time, now_time;
double now_close, now_open, now_low, now_high;
long now_volume;

int pre_time, last_fpos = 0;
double pre_close;

string prefix = "";
string symbolName;
bool justCreated;

bool Reset()
{
  ResetLastError();
  int deleted = CustomRatesDelete(symbolName, 0, LONG_MAX);
  int err = GetLastError();
  if(err != ERR_SUCCESS)
  {
    Alert("CustomRatesDelete failed, ", err);
    return false;
  }
  else
  {
    Print("Rates deleted: ", deleted);
  }

  ResetLastError();
  deleted = CustomTicksDelete(symbolName, 0, LONG_MAX);
  if(deleted == -1)
  {
    Print("CustomTicksDelete failed ", GetLastError());
    return false;
  }
  else
  {
    Print("Ticks deleted: ", deleted);
  }

  // wait for changes to take effect in background (asynchronously)
  int size;
  do
  {
    Sleep(1000);

    MqlTick _array[];
    size = CopyTicks(symbolName, _array, COPY_TICKS_ALL, 0, 10);
    Print("Remaining ticks: ", size);
  } while (size > 0);
  // NB.
  // still this can not work everytime as expected
  // if getting ERR_CUSTOM_TICKS_WRONG_ORDER or similar error - the last resort
  // is to wipe out the custom symbol manually from GUI, and then restart this EA

  return true; // success
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int OnInit()
{
  int _GetLastError = 0, cnt_bars = 0;
  justCreated = false;

  if(CustomPrefix != "")
  {
    prefix = CustomPrefix;
  }
  else
  {
    if(WorkMode == EqualVolumeBars)
      prefix = "_Eqv";
    else
      prefix = "_Rng";
  }

  // create custom symbol
  symbolName = prefix + _Symbol;
  if(!SymbolSelect(symbolName, true))
  {
    Print("< - - - Creating \"", symbolName, "\" - - - >");
    const SYMBOL Symb(symbolName);
    Symb.CloneProperties(_Symbol);
    justCreated = true;

    if(!SymbolSelect(symbolName, true))
    {
      Alert("Can't select symbol:", symbolName, " err:", GetLastError());
      return INIT_FAILED;
    }
  }
  else
  {
    Print("< - - - Resetting \"", prefix, _Symbol, "\" - - - >");
    if(!Reset()) return INIT_FAILED;
  }

  //+------------------------------------------------------------------+
  //| Process history
  //+------------------------------------------------------------------+
  ulong cursor = StartDate * 1000;

  double bid;
  now_time = 60;
  now_close = 0;
  now_open = 0;
  now_low = 0;
  now_high = 0;
  now_volume = 0;

  if(!FromM1)
  {
    Print("Processing tick history...");
    while(TicksBufferFilled(cursor))
    {
      MqlTick t;
      while(TicksBufferRead(t))
      {
        time = t.time;
        bid = t.bid;

        // New or first bar
        if(IsNewBar() || now_volume < 1)
        {
          if(IsNewBar())
          {
            WriteToFile(now_time, now_open, now_low, now_high, now_close, now_volume);
            cnt_bars++;
          }

          // Normalize down to a minute
          time = time / 60;
          time *= 60;

          // Eliminate 2 bars with the same name
          if(time <= now_time) time = now_time + 60;

          now_time = time;
          now_open = bid;
          now_low = bid;
          now_high = bid;
          now_close = bid;
          now_volume = 1;
        }
        else
        {
          if(bid < now_low) now_low = bid;
          if(bid > now_high) now_high = bid;
          now_close = bid;
          now_volume++;
        }
      }
    }
    Comment("");
  }
  // FromM1
  else
  {
    Print("Processing M1 bars history...");
    for(int i = Bars(_Symbol, PERIOD_M1) - 1; i >= 0; i--)
    {
      now_time = iTime(_Symbol, PERIOD_M1, i);
      // New or first bar
      if(IsNewBar() || now_volume < 1)
      {
        while(IsNewBar())
        {
          WriteToFile(now_time, now_open, now_low, now_high, now_close, TicksInBar);
          cnt_bars++;
          now_volume -= TicksInBar;
          now_time += 60;
        }

        time = iTime(_Symbol, PERIOD_M1, i);

        // Eliminate 2 bars with the same name
        while(time <= now_time) time = now_time + 60;

        now_time = time;
        now_open = iOpen(_Symbol, PERIOD_M1, i);
        now_low = iLow(_Symbol, PERIOD_M1, i);
        now_high = iHigh(_Symbol, PERIOD_M1, i);
        now_close = iClose(_Symbol, PERIOD_M1, i);
        now_volume += iVolume(_Symbol, PERIOD_M1, i);
      }
      else
      {
        if(iLow(_Symbol, PERIOD_M1, i) < now_low) now_low = iLow(_Symbol, PERIOD_M1, i);
        if(iHigh(_Symbol, PERIOD_M1, i) > now_high) now_high = iHigh(_Symbol, PERIOD_M1, i);
        now_close = iClose(_Symbol, PERIOD_M1, i);
        now_volume += iVolume(_Symbol, PERIOD_M1, i);
      }
    }
  }

  // write bar 0 to chart
  WriteToFile(now_time, now_open, now_low, now_high, now_close, now_volume);

  // show stats
  Print("< - - - ", cnt_bars, " bars written - - - >");
  Print("< - - - Open \"", prefix, _Symbol, "\" chart to view results - - - >");

  OpenCustomChart();

  RefreshWindow(now_time);

  return(INIT_SUCCEEDED);
}

MqlTick array[];
int tick;

bool TicksBufferFilled(ulong & cursor)
{
  Comment((datetime) cursor / 1000);
  int size = CopyTicks(_Symbol, array, COPY_TICKS_ALL, cursor, TICKS_ARRAY);
  if(size == -1)
  {
    Print("CopyTicks failed: ", GetLastError());
    return false;
  }
  else if(size == 0)
  {
    Print("End of CopyTicks at ", (datetime)(cursor / 1000));
    return false;
  }

  cursor = array[size - 1].time_msc + 1;
  tick = 0;

  return true;
}

bool TicksBufferRead(MqlTick & t)
{
  if(tick < ArraySize(array))
  {
    t = array[tick++];
    return true;
  }
  return false;
}

void OpenCustomChart()
{
  if(justCreated)
  {
    long id = ChartOpen(symbolName, PERIOD_M1);
    if(id == 0)
    {
      Alert("Can't open new chart for ", symbolName, ", code: ", GetLastError());
    }
    else
    {
      Sleep(1000);
      ChartSetSymbolPeriod(id, symbolName, PERIOD_M1);
      ChartSetInteger(id, CHART_MODE, CHART_CANDLES);
    }
    justCreated = false;
  }
}

void OnDeinit(const int reason)
{
}

void OnTick()
{
  //+------------------------------------------------------------------+
  //| Process incoming ticks
  //+------------------------------------------------------------------+
  MqlTick t;
  SymbolInfoTick(_Symbol, t);

  now_volume++;

  // bar continues
  if(IsOldBar())
  {
    if(t.bid < now_low) now_low = t.bid;
    if(t.bid > now_high) now_high = t.bid;
    now_close = t.bid;

    // write bar 0 to chart
    WriteToFile(now_time, now_open, now_low, now_high, now_close, now_volume);
  }
  // new bar tick
  else
  {
    // write bar 1
    WriteToFile(now_time, now_open, now_low, now_high, now_close, now_volume - 1);

    // normalize down to a minute
    time = iTime(_Symbol, PERIOD_M1, 0) / 60;
    time *= 60;

    // Eliminate 2 bars with the same name
    while(time <= now_time) time = now_time + 60;

    now_time = time;
    now_open = t.bid;
    now_low = t.bid;
    now_high = t.bid;
    now_close = t.bid;
    now_volume = 1;

    // write bar 0
    WriteToFile(now_time, now_open, now_low, now_high, now_close, now_volume);
  }

  RefreshWindow(now_time);
}

void RefreshWindow(const datetime t)
{
  MqlTick ta[1];
  SymbolInfoTick(_Symbol, ta[0]);
  ta[0].time = t;
  ta[0].time_msc = ta[0].time * 1000;
  if(CustomTicksAdd(symbolName, ta) == -1)
  {
    Print("Not ticked:", GetLastError(), " ", (long) ta[0].time);
    ArrayPrint(ta);
  }
  // remove the temporary tick
  CustomTicksDelete(symbolName, ta[0].time_msc, LONG_MAX);
}

void WriteToFile(datetime t, double o, double l, double h, double c, long v)
{
  MqlRates r[1];

  r[0].time = t;
  r[0].open = o;
  r[0].low = l;
  r[0].high = h;
  r[0].close = c;
  r[0].tick_volume = v;
  r[0].spread = 0;
  r[0].real_volume = 0;

  CustomRatesUpdate(symbolName, r);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool IsNewBar()
{
  if(WorkMode == EqualVolumeBars)
  {
    if(now_volume >= TicksInBar) return(true);
  }
  else
  {
    if((now_high - now_low) / _Point >= TicksInBar) return(true);
  }

  return(false);
}

bool IsOldBar()
{
  if(WorkMode == EqualVolumeBars)
  {
    if(now_volume <= TicksInBar) return(true);
  }
  else
  {
    if((now_high - now_low) / _Point <= TicksInBar) return(true);
  }

  return(false);
}